'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.LinearStringIndexMap = exports.LogStringIndexMap = exports.StringIndexMap = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }(); /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * Linear-time computation of the Longest Common Substring.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * Copyright (C) 2017 Kim Burgaard <kim@burgaard.us>
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * Permission is hereby granted, free of charge, to any person obtaining a copy
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * of this software and associated documentation files (the "Software"), to deal
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * in the Software without restriction, including without limitation the rights
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * copies of the Software, and to permit persons to whom the Software is
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * furnished to do so, subject to the following conditions:
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * The above copyright notice and this permission notice shall be included in
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * all copies or substantial portions of the Software.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * SOFTWARE.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      */

exports.default = longestCommonSubstring;

var _suffixArray = require('./suffix-array');

var _suffixArray2 = _interopRequireDefault(_suffixArray);

var _longestCommonPrefix = require('./longest-common-prefix');

var _longestCommonPrefix2 = _interopRequireDefault(_longestCommonPrefix);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Maps the position of strings s1 ... sK when concatenated into one string.
 * Concrete implementations provide different compromises between O(1) and
 * O(log(K)) lookup times versus O(n) and O(k) space requirements.
 */
var StringIndexMap = exports.StringIndexMap = function () {
  /**
   * Initializes a new string index map.
   * @param {number} k the number of expected strings.
   */
  function StringIndexMap() {
    var k = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0;

    _classCallCheck(this, StringIndexMap);

    if (typeof k !== 'number') {
      throw TypeError('k is not an integer: ' + (typeof k === 'undefined' ? 'undefined' : _typeof(k)));
    }
    if (k < 0) {
      throw Error('k may not be negative');
    }
    this.add = this.add.bind(this);
    this.lookup = this.lookup.bind(this);
    this.toString = this.toString.bind(this);
    this.ranges = new Array(k);
    this.k = 0;
    this.n = 0;
  }

  /* eslint-disable class-methods-use-this */

  /**
   * Adds a substring with the given length.
   * @throws {Error} not implemented.
   */


  _createClass(StringIndexMap, [{
    key: 'add',
    value: function add() {
      throw new Error('not implemented');
    }

    /**
     * Looks up the substring corresponding to the given position
     * in the concatenated string.
     * @throws {Error} not implemented.
     */

  }, {
    key: 'lookup',
    value: function lookup() {
      throw new Error('not implemented');
    }

    /**
     * Returns a string representation of the string index map.
     * @return {string} a string representation.
     */

  }, {
    key: 'toString',
    value: function toString() {
      throw new Error('not implemented');
    }
    /* eslint-enable class-methods-use-this */

  }]);

  return StringIndexMap;
}();

/**
 * Given a string of length n composed of substrings s1 ... sK, this string index map enables
 * O(log(K)) substring lookup in O(K) space. This implementation is preferable for small K and/or
 * very large n.
 */


var LogStringIndexMap = exports.LogStringIndexMap = function (_StringIndexMap) {
  _inherits(LogStringIndexMap, _StringIndexMap);

  function LogStringIndexMap() {
    _classCallCheck(this, LogStringIndexMap);

    return _possibleConstructorReturn(this, (LogStringIndexMap.__proto__ || Object.getPrototypeOf(LogStringIndexMap)).apply(this, arguments));
  }

  _createClass(LogStringIndexMap, [{
    key: 'add',

    /**
     * Adds a substring with the given length. O(1)
     * @param {number} length the length of the substring.
     * @return {number} the current total length of all substrings.
     */
    value: function add(length) {
      if (length < 1) {
        throw Error('the length must be a number greater than 0: ' + length);
      }
      var m = this.n + length;
      var k = this.k;
      this.ranges[this.k++] = [k, this.n];
      return this.n = m;
    }

    /**
     * Looks up the substring corresponding to the given position
     * in the concatenated string. O(log(K)) binary search.
     * @param {number} position the position in the concatenated string.
     * @return {number} the index of the substring that contains the given position.
     */

  }, {
    key: 'lookup',
    value: function lookup(suffix, start, end) {
      if (start < 0) {
        throw new Error('start ' + start + ' is less than 0');
      }
      if (end <= start) {
        throw new Error('end ' + end + ' is less than or equal to start ' + start);
      }

      var s = start || 0;
      var e = end || this.ranges.length;
      var n = e - s;
      if (n === 1) {
        return this.ranges[s][0];
      }

      var m = s + (n >>> 1); // eslint-disable-line no-bitwise
      return suffix < this.ranges[m][1] ? this.lookup(suffix, s, m) : this.lookup(suffix, m, e);
    }
  }, {
    key: 'toString',
    value: function toString() {
      return '[' + this.ranges.map(function (r) {
        return '[' + r.join(', ') + ']';
      }).join(', ') + ']';
    }
  }]);

  return LogStringIndexMap;
}(StringIndexMap);

/**
 * Given a string of length n composed of substrings s1 ... sK, this string index map enables
 * log(1) substring lookup in O(n) space. This implementation is preferable for very large K and/or
 * small n.
 */


var LinearStringIndexMap = exports.LinearStringIndexMap = function (_StringIndexMap2) {
  _inherits(LinearStringIndexMap, _StringIndexMap2);

  function LinearStringIndexMap() {
    _classCallCheck(this, LinearStringIndexMap);

    var _this2 = _possibleConstructorReturn(this, (LinearStringIndexMap.__proto__ || Object.getPrototypeOf(LinearStringIndexMap)).call(this));

    _this2.initialize = _this2.initialize.bind(_this2);
    _this2.initialized = false;
    return _this2;
  }

  /**
   * Adds a substring with the given length. O(1)
   * @param {number} length the length of the substring.
   * @return {number} the current total length of all substrings.
   */


  _createClass(LinearStringIndexMap, [{
    key: 'add',
    value: function add(length) {
      if (length < 1) {
        throw Error('the length must be a number greater than 0: ' + length);
      }
      this.initialized = false;
      this.ranges[this.k++] = length;
      return this.n += length;
    }

    /**
     * Fills in the index array in O(n).
     * @return {Boolean} true if the array was initialized, false if no initialization was needed.
     */

  }, {
    key: 'initialize',
    value: function initialize() {
      var _this3 = this;

      if (this.initialized) {
        return false;
      }

      this.indexMap = new Array(this.n);

      var i = 0;
      this.ranges.forEach(function (r, j) {
        _this3.indexMap.fill(j, i, i + r);
        i += r;
      });

      return this.initialized = true;
    }

    /**
     * Looks up the substring corresponding to the given position
     * in the concatenated string. O(1) array lookup [O(n) on first lookup to initialize index map]
     * @param {number} position the position in the concatenated string.
     * @return {number} the index of the substring that contains the given position.
     */

  }, {
    key: 'lookup',
    value: function lookup(suffix) {
      this.initialize();

      return this.indexMap[suffix];
    }
  }, {
    key: 'toString',
    value: function toString() {
      this.initialize();

      return '{\n  ranges: [' + this.ranges.join(', ') + '],\n  indexMap: [' + this.indexMap.join(', ') + ']\n}';
    }
  }]);

  return LinearStringIndexMap;
}(StringIndexMap);

function createStringIndexMap(indexMap, k) {
  if (indexMap instanceof StringIndexMap) {
    return indexMap;
  }

  if (typeof indexMap !== 'string') {
    throw new TypeError('the index map argument must be a string or a StringIndexMap instance: ' + (typeof indexMap === 'undefined' ? 'undefined' : _typeof(indexMap)));
  }

  switch (indexMap) {
    case 'log':
      return new LogStringIndexMap(k);
    case 'linear':
      return new LinearStringIndexMap(k);
    default:
      throw new Error('strategy must be either \'log\' or \'linear\': ' + indexMap);
  }
}

function ensureNonEmptyString(e) {
  if (typeof e !== 'string') {
    throw new TypeError('entry is not a string: ' + (typeof e === 'undefined' ? 'undefined' : _typeof(e)));
  }
  if (e.length === 0) {
    throw new Error('entry is an empty string');
  }
}

/**
 * Finds the longest common substring(s) in the set of given strings. If there
 * are multiple substrings that all share the longest length, then all such
 * substrings are returned. O(n) or O(n * log(K)) depending on the selected
 * string indexing strategy.
 * 
 * @param {string[]} strings an array of strings.
 * @param {string|StringIndexMap} [indexMap='log'] string indexing map. If
 *   given a string, it must be one of 'log' or 'linear'.
 * @return {string|string[]} the longest common substring(s).
 */
function longestCommonSubstring(strings) {
  var indexMap = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'log';

  if (!Array.isArray(strings)) {
    throw new TypeError('strings argument must be an array of strings');
  }

  var k = strings.length;
  // handle trivial cases
  switch (k) {
    case 0:
      return [];
    case 1:
      ensureNonEmptyString(strings[0]);
      return strings;
    default:
      break;
  }

  var stringIndexMap = createStringIndexMap(indexMap, k);

  // append each string as an array of unicodes + unique terminator
  // O(n)
  var terminator = -1;
  var sequence = strings.reduce(function (a, e) {
    ensureNonEmptyString(e);
    stringIndexMap.add(e.length + 1);
    return a.concat(e.split('').map(function (c) {
      return c.charCodeAt(0);
    }).concat(terminator--));
  }, []);

  // calculate the suffix array in O(n)
  var sa = (0, _suffixArray2.default)(sequence, terminator);

  // calculate the longest common prefixes in O(n)
  var lcp = (0, _longestCommonPrefix2.default)(sequence.concat(terminator), sa);

  var result = [];
  var longest = 1;
  var top = -1;
  var suffix = sa[0];
  var stringIndex = stringIndexMap.lookup(sa[0]);
  var entryStack = [];

  var i = 0;
  while (i < sa.length - 1) {
    var h = lcp[i];
    var suffixNext = sa[i + 1];
    var stringIndexNext = stringIndexMap.lookup(suffixNext);
    var j = 0;
    if (h >= top && h > 0) {
      // add a stack entry for each height level and map the string index to its suffix and length
      while (j < h) {
        if (j >= entryStack.length) {
          entryStack[j] = {};
        }
        entryStack[j][stringIndex] = [suffix, j + 1];
        entryStack[j][stringIndexNext] = [suffixNext, j + 1];
        j++;
      }
      top = j;
    }

    // if the height goes down, then check what we have on the stack
    j = top;
    while (j > h) {
      var entries = entryStack.pop();
      if (j >= longest) {
        var keys = Object.keys(entries);
        // check if all k strings have been seen
        if (keys.length === k) {
          var r = entries[keys[0]];
          if (j > longest) {
            // reset result for the longest substring seen so far
            longest = j;
            result = [r];
          } else {
            // add substring to its peers of longest substrings seen so far
            result.push(r);
          }
        }
      }
      j--;
    }

    top = h;
    i++;
    suffix = suffixNext;
    stringIndex = stringIndexNext;
  }

  // check if we have any remaining entries on the stack
  // TODO refactor and remove duplicated code
  while (entryStack.length > 0) {
    var _j = entryStack.length;
    var _entries = entryStack.pop();
    if (_j >= longest) {
      var _keys = Object.keys(_entries);
      if (_keys.length === k) {
        var _r = _entries[_keys[0]];
        if (_j > longest) {
          longest = _j;
          result = [_r];
        } else {
          result.push(_r);
        }
      }
    }
  }

  return result.map(function (r) {
    return sequence.slice(r[0], r[0] + r[1]).map(function (c) {
      return String.fromCharCode(c);
    }).join('');
  });
}